#include "CueStick.h"

CueStick::CueStick(SDL_Surface *_screen, PoolGame *_game)
	: screen(_screen), game(_game), distance(10), angle(0),
	  visible(false), targetBall(NULL), listener(NULL)
{
	stickLength = 60;
	tipLength   = 6;

	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, WIDTH,
								   HEIGHT, 16, 0, 0, 0, 0);

	for (int i = 0; i < 3; i++)
	{
		oldX[i] = x[i] = 0;
		oldY[i] = y[i] = 0;
	}
}

CueStick::~CueStick()
{
}

const char *
CueStick::getName() const
{
	return "CueStick";
}

void
CueStick::setVisible(bool visible)
{
	if (this->visible == visible)
		return;

	this->visible = visible;

	if (visible == false)
	{
		SDL_Rect rect;

		Sint16 lX, tY, rX, bY;

		/* Start these off with something. */
		lX = x[0];
		tY = y[0];
		rX = x[0];
		bY = y[0];

		/* Get the bounding rect. */
		for (int i = 0; i < 3; i++)
		{
			lX = MIN(lX, x[i]);
			rX = MAX(rX, x[i]);
			tY = MIN(tY, y[i]);
			bY = MAX(bY, y[i]);
		}

		rect.x = lX;
		rect.y = tY;
		rect.w = rX - lX + 1;
		rect.h = bY - tY + 1;

		/* Fill in the background. */
		game->getRoom()->render(rect, true);
	}
	else if (x[0] != 0)
		render();
}

bool
CueStick::isVisible() const
{
	return visible;
}

void
CueStick::setTargetBall(Ball *ball)
{
	targetBall = ball;
	distance = ball->getRadius() + 2;
}

Ball *
CueStick::getTargetBall() const
{
	return targetBall;
}

void
CueStick::rotateLeft(Fixed degrees)
{
	setAngle(getAngleDeg() + degrees);
}

void
CueStick::rotateRight(Fixed degrees)
{
	setAngle(getAngleDeg() - degrees);
}

void
CueStick::setAngle(Fixed angle)
{
	setAngleRad(angle * M_PI / 180.0);
}

void
CueStick::setAngleRad(Fixed angle)
{
	Fixed oldAngle = getAngleDeg();

	/* Convert from degrees to radians */
	this->angle = angle;

	if (listener != NULL)
		listener->stickMoved(this, oldAngle, getAngleDeg());
}

Fixed
CueStick::getAngleRad() const
{
	return angle;
}

Fixed
CueStick::getAngleDeg() const
{
	return (angle * 180.0) / M_PI;
}

void
CueStick::setDistance(Fixed distance)
{
	this->distance = distance;
}

Fixed
CueStick::getDistance() const
{
	return distance;
}

bool
CueStick::hitBall(Fixed level, Fixed maxLevel)
{
	bool result = false;

	Fixed power = level / maxLevel;

	setVisible(false);

	game->getPhysics()->shootBall(getTargetBall(), power, angle);

	if (listener != NULL)
		result = listener->cueBallHit(this, getTargetBall());

	setVisible(true);

	return result;
}

void
CueStick::registerListener(CueStickListener *listener)
{
	this->listener = listener;
}

void
CueStick::render(bool update)
{
	if (!isVisible())
		return;

	Fixed centerX   = targetBall->getCenterX();
	Fixed centerY   = targetBall->getCenterY();
	Fixed distance  = getDistance();
	Fixed baseWidth = 4;
	Fixed cosangle  = angle.cos();
	Fixed sinangle  = angle.sin();
	Fixed a, b, c;
	Sint16 tipX[3], tipY[3];
	Sint16 lX, tY, rX, bY;
	SDL_Rect rect;
	int i;

	/* Save the x, y coords. */
	for (i = 0; i < 3; i++)
	{
		oldX[i] = x[i];
		oldY[i] = y[i];
	}

	Fixed _x = baseWidth / 2.0;
	Fixed sindist = sinangle * distance;
	Fixed cosdist = cosangle * distance;

	/* Set the points for the body of the cue stick */
	Fixed negStickDist = (distance + stickLength).neg();

	a = sinangle * negStickDist;
	b = cosangle * negStickDist;

	x[0] = (centerX + sindist).intVal();
	y[0] = (centerY - cosdist).intVal();

	x[1] = (centerX + cosangle * _x       - a).intVal();
	y[1] = (centerY - sinangle * _x.neg() + b).intVal();

	x[2] = (centerX + cosangle * _x.neg() - a).intVal();
	y[2] = (centerY - sinangle * _x       + b).intVal();

	/* Set the points for the tip of the stick. */
	a = (tipLength.neg() / (stickLength.neg() / _x));
	b = (distance + tipLength).neg();
	c = (tipLength.neg() / (stickLength.neg() / _x.neg()));

	tipX[0] = (centerX + sindist).intVal();
	tipY[0] = (centerY - cosdist).intVal();

	tipX[1] = (centerX + cosangle * a       - sinangle * b).intVal();
	tipY[1] = (centerY - sinangle * a.neg() + cosangle * b).intVal();

	tipX[2] = (centerX + cosangle * c.neg() - sinangle * b).intVal();
	tipY[2] = (centerY - sinangle * c       + cosangle * b).intVal();
	
	/* Start these off with something. */
	lX = x[0];
	tY = y[0];
	rX = x[0];
	bY = y[0];

	/* Get the bounding rect. */
	for (i = 0; i < 3; i++)
	{
		lX = MIN(lX, x[i]);
		rX = MAX(rX, x[i]);
		tY = MIN(tY, y[i]);
		bY = MAX(bY, y[i]);

		if (oldX[i] != 0 || oldY[i] != 0)
		{
			lX = MIN(lX, oldX[i]);
			rX = MAX(rX, oldX[i]);
			tY = MIN(tY, oldY[i]);
			bY = MAX(bY, oldY[i]);
		}
	}

	rect.x = lX;
	rect.y = tY;
	rect.w = rX - lX + 1;
	rect.h = bY - tY + 1;

	/* Check the bounds. */
	if (rect.x < 0)
		rect.x = 0;
	else if (rect.x > WIDTH)
		rect.x = WIDTH;

	if (rect.y < 0)
		rect.y = 0;
	else if (rect.y > HEIGHT)
		rect.y = HEIGHT;

	if (rect.x + rect.w > WIDTH)
		rect.w = WIDTH - rect.x;

	if (rect.y + rect.h > HEIGHT)
		rect.h = HEIGHT - rect.y;

	/* Fill in the background. */
	game->getRoom()->render(rect, false);

	/* Render the balls so it doesn't disappear on us :P */
	game->getTable()->renderBalls();

	/* Draw the stick. */
	filledPolygonRGBA(screen, x, y, 3, 120, 64, 0, 0xFF);
	filledPolygonRGBA(screen, tipX, tipY, 3, 0xFF, 0xFF, 0xFF, 0xFF);

	if (update)
		SDL_UpdateRect(screen, rect.x, rect.y, rect.w, rect.h);
}

bool
CueStick::mousePressed(unsigned int x, unsigned int y)
{
	return mouseDragged(x, y);
}

bool
CueStick::mouseReleased(unsigned int, unsigned int)
{
	return false;
}

bool
CueStick::mouseDragged(unsigned int x, unsigned int y)
{
	Ball *ball = getTargetBall();

	setAngleRad(((ball->getCenterX() - x).neg() /
				 (ball->getCenterY() - y)).atan() +
				(ball->getCenterY() <= y ? M_PI : 0));

	return true;
}
