/*
 * libmad - MPEG audio decoder library
 * Copyright (C) 2000-2001 Robert Leslie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Fixed.h,v 1.4 2002/09/12 06:00:05 chipx86 Exp $
 */

# ifndef _FIXED_H
# define _FIXED_H

#define SIZEOF_INT 4

# if SIZEOF_INT >= 4
typedef   signed int fixed_t;

typedef   signed int fixed64hi_t;
typedef unsigned int fixed64lo_t;
# else
typedef   signed long fixed_t;

typedef   signed long fixed64hi_t;
typedef unsigned long fixed64lo_t;
# endif

# if defined(_MSC_VER)
#  define fixed64_t  signed __int64
# elif 1 || defined(__GNUC__)
#  define fixed64_t  signed long long
# endif

# if defined(FPM_FLOAT)
typedef double mad_sample_t;
# else
typedef fixed_t mad_sample_t;
# endif

/*
 * Fixed-point format: 0xABBBBBBB
 * A == whole part      (sign + 3 bits)
 * B == fractional part (28 bits)
 *
 * Values are signed two's complement, so the effective range is:
 * 0x80000000 to 0x7fffffff
 *       -8.0 to +7.9999999962747097015380859375
 *
 * The smallest representable value is:
 * 0x00000001 == 0.0000000037252902984619140625 (i.e. about 3.725e-9)
 *
 * 28 bits of fractional accuracy represent about
 * 8.6 digits of decimal accuracy.
 *
 * Fixed-point numbers can be added or subtracted as normal
 * integers, but multiplication requires shifting the 64-bit result
 * from 56 fractional bits back to 28 (and rounding.)
 *
 * Changing the definition of F_FRACBITS is only partially
 * supported, and must be done with care.
 */

# define F_FRACBITS		16

# if F_FRACBITS == 28
#  define MAD_F(x)		((fixed_t) (x##L))
# else
#  if F_FRACBITS < 28
//#   warning "F_FRACBITS < 28"
#   define MAD_F(x)		((fixed_t)  \
				 (((x##L) +  \
				   (1L << (28 - F_FRACBITS - 1))) >>  \
				  (28 - F_FRACBITS)))

#  elif F_FRACBITS > 28
#   error "F_FRACBITS > 28 not currently supported"
#   define MAD_F(x)		((fixed_t)  \
				 ((x##L) << (F_FRACBITS - 28)))
#  endif
# endif

# define F_ONE		MAD_F(0x10000000)
# define F_MIN		((fixed_t) -0x80000000L)
# define F_MAX		((fixed_t) +0x7FFFFFFEL)
# define F_NAN      ((fixed_t) +0x7FFFFFFFL)

# define F_FIXED(x)	((fixed_t)  \
				 ((x) * (double) (1L << F_FRACBITS) + 0.5))
# define F_DOUBLE(x)	((double)  \
				 ((x) / (double) (1L << F_FRACBITS)))

# define F_INT(x)	((x) >> F_FRACBITS)
# define F_FRACT(x)	((x) & ((1L << F_FRACBITS) - 1))
				/* (x should be positive) */

# define F_FROMINT(x)	((x) << F_FRACBITS)

# define F_ADD(x, y)	((x) + (y))
# define F_SUB(x, y)	((x) - (y))
# define F_DIV(x, y)    ((x) / (y))

# if defined(FPM_FLOAT)
#  error "FPM_FLOAT not yet supported"

#  undef MAD_F
#  define MAD_F(x)		F_DOUBLE(x)

#  define F_MUL(x, y)	((x) * (y))
#  define F_SCALE64

#  undef ASO_ZEROCHECK

# elif defined(FPM_64BIT)

/*
 * This version should be the most accurate if 64-bit types are supported by
 * the compiler, although it may not be the most efficient.
 */
#  if defined(OPT_ACCURACY)
#   define F_MUL(x, y)  \
    ((fixed_t)  \
     ((((fixed64_t) (x) * (y)) +  \
       (1L << (F_SCALEBITS - 1))) >> F_SCALEBITS))
#  else
#   define F_MUL(x, y)  \
    ((fixed_t) (((fixed64_t) (x) * (y)) >> F_SCALEBITS))
#  endif

#  define F_SCALEBITS  F_FRACBITS

/* --- Intel --------------------------------------------------------------- */

# elif defined(FPM_INTEL)

#  if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable: 4035)  /* no return value */
static __forceinline
fixed_t F_MUL_inline(fixed_t x, fixed_t y)
{
  enum {
    fracbits = F_FRACBITS
  };

  __asm {
    mov eax, x
    imul y
    shrd eax, edx, fracbits
  }

  /* implicit return of eax */
}
#   pragma warning(pop)

#   define F_MUL		F_MUL_inline
#   define F_SCALE64
#  else
/*
 * This Intel version is fast and accurate; the disposition of the least
 * significant bit depends on OPT_ACCURACY via F_SCALE64().
 */
#   define F_MLX(hi, lo, x, y)  \
    asm ("imull %3"  \
	 : "=a" (lo), "=d" (hi)  \
	 : "%a" (x), "rm" (y)  \
	 : "cc")

#   if defined(OPT_ACCURACY)
/*
 * This gives best accuracy but is not very fast.
 */
#    define F_MLA(hi, lo, x, y)  \
    ({ fixed64hi_t __hi;  \
       fixed64lo_t __lo;  \
       F_MLX(__hi, __lo, (x), (y));  \
       asm ("addl %2,%0\n\t"  \
	    "adcl %3,%1"  \
	    : "=rm" (lo), "=rm" (hi)  \
	    : "r" (__lo), "r" (__hi), "0" (lo), "1" (hi)  \
	    : "cc");  \
    })
#   endif  /* OPT_ACCURACY */

#   if defined(OPT_ACCURACY)
/*
 * Surprisingly, this is faster than SHRD followed by ADC.
 */
#    define F_SCALE64(hi, lo)  \
    ({ fixed64hi_t __hi_;  \
       fixed64lo_t __lo_;  \
       fixed_t __result;  \
       asm ("addl %4,%2\n\t"  \
	    "adcl %5,%3"  \
	    : "=rm" (__lo_), "=rm" (__hi_)  \
	    : "0" (lo), "1" (hi),  \
	      "ir" (1L << (F_SCALEBITS - 1)), "ir" (0)  \
	    : "cc");  \
       asm ("shrdl %3,%2,%1"  \
	    : "=rm" (__result)  \
	    : "0" (__lo_), "r" (__hi_), "I" (F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })
#   else
#    define F_SCALE64(hi, lo)  \
    ({ fixed_t __result;  \
       asm ("shrdl %3,%2,%1"  \
	    : "=rm" (__result)  \
	    : "0" (lo), "r" (hi), "I" (F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })
#   endif  /* OPT_ACCURACY */

#   define F_SCALEBITS  F_FRACBITS
#  endif

/* --- ARM ----------------------------------------------------------------- */

# elif defined(FPM_ARM)

/* 
 * This ARM V4 version is as accurate as FPM_64BIT but much faster. The
 * least significant bit is properly rounded at no CPU cycle cost!
 */
# if 1
/*
 * There's a bug somewhere, possibly in the compiler, that sometimes makes
 * this necessary instead of the default implementation via F_MLX and
 * F_SCALE64. It may be related to the use (or lack) of
 * -finline-functions and/or -fstrength-reduce.
 *
 * This is also apparently faster than F_MLX/F_SCALE64.
 */
#  define F_MUL(x, y)  \
    ({ fixed64hi_t __hi;  \
       fixed64lo_t __lo;  \
       fixed_t __result;  \
       asm ("smull	%0, %1, %3, %4\n\t"  \
	    "movs	%0, %0, lsr %5\n\t"  \
	    "adc	%2, %0, %1, lsl %6"  \
	    : "=&r" (__lo), "=&r" (__hi), "=r" (__result)  \
	    : "%r" (x), "r" (y),  \
	      "M" (F_SCALEBITS), "M" (32 - F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })
# endif

#  define F_MLX(hi, lo, x, y)  \
    asm ("smull	%0, %1, %2, %3"  \
	 : "=&r" (lo), "=&r" (hi)  \
	 : "%r" (x), "r" (y))

#  define F_MLA(hi, lo, x, y)  \
    asm ("smlal	%0, %1, %2, %3"  \
	 : "+r" (lo), "+r" (hi)  \
	 : "%r" (x), "r" (y))

#  define F_MLN(hi, lo)  \
    asm ("rsbs	%0, %2, #0\n\t"  \
	 "rsc	%1, %3, #0"  \
	 : "=r" (lo), "=r" (hi)  \
	 : "0" (lo), "1" (hi)  \
	 : "cc")

#  define F_SCALE64(hi, lo)  \
    ({ fixed_t __result;  \
       asm ("movs	%0, %1, lsr %3\n\t"  \
	    "adc	%0, %0, %2, lsl %4"  \
	    : "=r" (__result)  \
	    : "r" (lo), "r" (hi),  \
	      "M" (F_SCALEBITS), "M" (32 - F_SCALEBITS)  \
	    : "cc");  \
       __result;  \
    })

#  define F_SCALEBITS  F_FRACBITS

/* --- MIPS ---------------------------------------------------------------- */

# elif defined(FPM_MIPS)

/*
 * This MIPS version is fast and accurate; the disposition of the least
 * significant bit depends on OPT_ACCURACY via F_SCALE64().
 */
#  define F_MLX(hi, lo, x, y)  \
    asm ("mult	%2,%3"  \
	 : "=l" (lo), "=h" (hi)  \
	 : "%r" (x), "r" (y))

# if defined(HAVE_MADD_ASM)
#  define F_MLA(hi, lo, x, y)  \
    asm ("madd	%2,%3"  \
	 : "+l" (lo), "+h" (hi)  \
	 : "%r" (x), "r" (y))
# elif defined(HAVE_MADD16_ASM)
/*
 * This loses significant accuracy due to the 16-bit integer limit in the
 * multiply/accumulate instruction.
 */
#  define F_ML0(hi, lo, x, y)  \
    asm ("mult	%2,%3"  \
	 : "=l" (lo), "=h" (hi)  \
	 : "%r" ((x) >> 12), "r" ((y) >> 16))
#  define F_MLA(hi, lo, x, y)  \
    asm ("madd16	%2,%3"  \
	 : "+l" (lo), "+h" (hi)  \
	 : "%r" ((x) >> 12), "r" ((y) >> 16))
#  define F_MLZ(hi, lo)  ((fixed_t) (lo))
# endif

# if defined(OPT_SPEED)
#  define F_SCALE64(hi, lo)  \
    ((fixed_t) ((hi) << (32 - F_SCALEBITS)))
#  define F_SCALEBITS  F_FRACBITS
# endif

/* --- SPARC --------------------------------------------------------------- */

# elif defined(FPM_SPARC)

/*
 * This SPARC V8 version is fast and accurate; the disposition of the least
 * significant bit depends on OPT_ACCURACY via F_SCALE64().
 */
#  define F_MLX(hi, lo, x, y)  \
    asm ("smul %2, %3, %0\n\t"  \
	 "rd %%y, %1"  \
	 : "=r" (lo), "=r" (hi)  \
	 : "%r" (x), "rI" (y))

/* --- PowerPC ------------------------------------------------------------- */

# elif defined(FPM_PPC)

/*
 * This PowerPC version is tuned for the 4xx embedded processors. It is
 * effectively a tuned version of FPM_64BIT. It is a little faster and just
 * as accurate. The disposition of the least significant bit depends on
 * OPT_ACCURACY via F_SCALE64().
 */
#  define F_MLX(hi, lo, x, y)  \
    asm ("mulhw %1, %2, %3\n\t"  \
	 "mullw %0, %2, %3"  \
	 : "=&r" (lo), "=&r" (hi)  \
	 : "%r" (x), "r" (y))

#  define F_MLA(hi, lo, x, y)  \
    ({ fixed64hi_t __hi;  \
       fixed64lo_t __lo;  \
       F_MLX(__hi, __lo, (x), (y));  \
       asm ("addc %0, %2, %3\n\t"  \
	    "adde %1, %4, %5"  \
	    : "=r" (lo), "=r" (hi)  \
	    : "%r" (__lo), "0" (lo), "%r" (__hi), "1" (hi));  \
    })

#  if defined(OPT_ACCURACY)
/*
 * This is accurate and ~2 - 2.5 times slower than the unrounded version.
 *
 * The __volatile__ improves the generated code by another 5% (fewer spills
 * to memory); eventually they should be removed.
 */
#   define F_SCALE64(hi, lo)  \
    ({ fixed_t __result;  \
       fixed64hi_t __hi_;  \
       fixed64lo_t __lo_;  \
       asm __volatile__ ("addc %0, %2, %4\n\t"  \
			 "addze %1, %3"  \
	    : "=r" (__lo_), "=r" (__hi_)  \
	    : "r" (lo), "r" (hi), "r" (1 << (F_SCALEBITS - 1)));  \
       asm __volatile__ ("rlwinm %0, %2,32-%3,0,%3-1\n\t"  \
			 "rlwimi %0, %1,32-%3,%3,31"  \
	    : "=&r" (__result)  \
	    : "r" (__lo_), "r" (__hi_), "I" (F_SCALEBITS));  \
	    __result;  \
    })
#  else
#   define F_SCALE64(hi, lo)  \
    ({ fixed_t __result;  \
       asm ("rlwinm %0, %2,32-%3,0,%3-1\n\t"  \
	    "rlwimi %0, %1,32-%3,%3,31"  \
	    : "=r" (__result)  \
	    : "r" (lo), "r" (hi), "I" (F_SCALEBITS));  \
	    __result;  \
    })
#  endif  /* OPT_ACCURACY */

#  define F_SCALEBITS  F_FRACBITS

/* --- Default ------------------------------------------------------------- */

# elif defined(FPM_DEFAULT)

/*
 * This version is the most portable but it loses significant accuracy.
 * Furthermore, accuracy is biased against the second argument, so care
 * should be taken when ordering operands.
 *
 * The scale factors are constant as this is not used with SSO.
 *
 * Pre-rounding is required to stay within the limits of compliance.
 */
#  if defined(OPT_SPEED)
#   define F_MUL(x, y)	(((x) >> 12) * ((y) >> 16))
#  else
#   define F_MUL(x, y)	((((x) + (1L << 11)) >> 12) *  \
				 (((y) + (1L << 15)) >> 16))
#  endif

/* ------------------------------------------------------------------------- */

# else
#  error "no FPM selected"
# endif

/* default implementations */

# if !defined(F_MUL)
#  define F_MUL(x, y)  \
    ({ fixed64hi_t __hi;  \
       fixed64lo_t __lo;  \
       F_MLX(__hi, __lo, (x), (y));  \
       F_SCALE64(__hi, __lo);  \
    })
# endif

# if !defined(F_MLA)
#  define F_ML0(hi, lo, x, y)	((lo)  = F_MUL((x), (y)))
#  define F_MLA(hi, lo, x, y)	((lo) += F_MUL((x), (y)))
#  define F_MLN(hi, lo)		((lo)  = -(lo))
#  define F_MLZ(hi, lo)		((void) (hi), (fixed_t) (lo))
# endif

# if !defined(F_ML0)
#  define F_ML0(hi, lo, x, y)	F_MLX((hi), (lo), (x), (y))
# endif

# if !defined(F_MLN)
#  define F_MLN(hi, lo)		((hi) = ((lo) = -(lo)) ? ~(hi) : -(hi))
# endif

# if !defined(F_MLZ)
#  define F_MLZ(hi, lo)		F_SCALE64((hi), (lo))
# endif

# if !defined(F_SCALE64)
#  if defined(OPT_ACCURACY)
#   define F_SCALE64(hi, lo)  \
    ((((fixed_t)  \
       (((hi) << (32 - (F_SCALEBITS - 1))) |  \
	((lo) >> (F_SCALEBITS - 1)))) + 1) >> 1)
#  else
#   define F_SCALE64(hi, lo)  \
    ((fixed_t)  \
     (((hi) << (32 - F_SCALEBITS)) |  \
      ((lo) >> F_SCALEBITS)))
#  endif
#  define F_SCALEBITS  F_FRACBITS
# endif

class Fixed
{
	public:
		Fixed() : val(0) {}
		Fixed(double num);
		Fixed(int num);
		Fixed(unsigned int num);
		~Fixed();

		Fixed operator=(const Fixed &f);
		
		bool operator==(const Fixed &f) const;
		bool operator!=(const Fixed &f) const;
		bool operator<(const Fixed &f)  const;
		bool operator<=(const Fixed &f) const;
		bool operator>(const Fixed &f)  const;
		bool operator>=(const Fixed &f) const;

		Fixed operator+(const Fixed &f) const;
		Fixed operator-(const Fixed &f) const;
		Fixed operator*(const Fixed &f) const;
		Fixed operator/(const Fixed &f) const;

		void operator+=(const Fixed &f);
		void operator-=(const Fixed &f);
		void operator*=(const Fixed &f);
		void operator/=(const Fixed &f);

		Fixed set(fixed_t fixed);

		double doubleVal() const;
		unsigned int intVal() const;
		int sintVal() const;
		fixed_t value() const;

		bool isNaN() const;
		
		Fixed abs() const;
		Fixed sqrt() const;
		Fixed hypot(const Fixed &f) const;
		Fixed cos() const;
		Fixed sin() const;
		Fixed atan() const;

		Fixed square() const;
		Fixed neg() const;

	private:
//		Fixed(fixed_t num);

		fixed_t val;

		bool nan;
};

#endif
