#include "PoolGame.h"

PoolGame::PoolGame(SDL_Surface *screen)
	: curPlayer(0), numPlayers(0), winner(NULL), playing(false)
{
	room    = new PoolRoom(screen, this);
	physics = new Physics();
	rules   = new EightBallRules(this);

	players[0] = NULL;
	players[1] = NULL;

	addPlayer(new Player(0));
	addPlayer(new Player(1));

	mouseListeners.append(room->getPowerMeter());
}

PoolGame::~PoolGame()
{
	delete players[0];

	if (!isSinglePlayer())
		delete players[1];

	delete rules;
	delete room;
}

void
PoolGame::start()
{
	playing = true;

	getRules()->rackBalls();
	getTable()->resetCueBall();
	getRoom()->render();

	eventLoop();
}

void
PoolGame::end()
{
	playing = false;
}

void
PoolGame::setRules(Rules *rules)
{
	if (this->rules != NULL)
		delete this->rules;

	this->rules = rules;
}

PoolRoom *
PoolGame::getRoom() const
{
	return room;
}

PoolTable *
PoolGame::getTable() const
{
	return room->getTable();
}

Physics *
PoolGame::getPhysics() const
{
	return physics;
}

Rules *
PoolGame::getRules() const
{
	return rules;
}

void
PoolGame::addPlayer(Player *player)
{
	if (numPlayers < 2)
		players[numPlayers++] = player;
}

void
PoolGame::setWinner(Player *player)
{
	winner = player;
}

Player *
PoolGame::getActivePlayer() const
{
	return players[curPlayer];
}

Player *
PoolGame::getPlayer(int num) const
{
	return players[num <= 0 ? 0 : 1];
}

Player *
PoolGame::getWinner() const
{
	return winner;
}

int
PoolGame::getPlayerCount() const
{
	return numPlayers;
}

Player *
PoolGame::getInactivePlayer() const
{
	if (isSinglePlayer())
		return NULL;

	return players[curPlayer == 0 ? 1 : 0];
}

bool
PoolGame::isSinglePlayer() const
{
	return (numPlayers == 1);
}

void
PoolGame::switchPlayers()
{
	if (isSinglePlayer())
		return;

	curPlayer = (curPlayer == 0 ? 1 : 0);

	getRoom()->getPlayerDisplay()->render();
}

bool
PoolGame::isPlaying() const
{
	return playing;
}

void
PoolGame::eventLoop()
{
	SDL_Event event;
	MouseListener *mouseTarget = NULL;

	while (isPlaying())
	{
		while (SDL_PollEvent(&event))
		{
			if (event.type == SDL_KEYDOWN)
			{
				switch (event.key.keysym.sym)
				{
					case SDLK_ESCAPE:
						end();
						break;

					case SDLK_SPACE:
					{
						PowerMeter *powerMeter = getRoom()->getPowerMeter();

						if (powerMeter->getLevel() > 0 &&
							!room->getCueStick()->hitBall(
								powerMeter->getLevel(),
								powerMeter->getMaxLevel()))
						{
							switchPlayers();
						}

						break;
					}

					case SDLK_LEFT:
						room->getCueStick()->rotateLeft(5);
						break;

					case SDLK_RIGHT:
						room->getCueStick()->rotateRight(5);
						break;

					default:
						break;
				}
			}
			else if (event.type == SDL_MOUSEBUTTONDOWN)
			{
				Uint16 x = event.button.x;
				Uint16 y = event.button.y;
				Sprite *listener;

				for (listener = mouseListeners.first();
					 listener != NULL;
					 listener = mouseListeners.next())
				{
					if (listener->containsPoint(x, y))
					{
						mouseTarget = listener;
						break;
					}
				}

				/* The cue stick is a special exception for now. */
				if (mouseTarget == NULL)
					mouseTarget = getRoom()->getCueStick();

				mouseTarget->mousePressed(x, y);
			}
			else if (event.type == SDL_MOUSEBUTTONUP)
			{
				Uint16 x = event.button.x;
				Uint16 y = event.button.y;

				if (mouseTarget != NULL)
				{
					mouseTarget->mouseReleased(x, y);

					mouseTarget = NULL;
				}
			}
			else if (event.type == SDL_MOUSEMOTION)
			{
				Uint16 x = event.button.x;
				Uint16 y = event.button.y;

				if (mouseTarget != NULL)
					mouseTarget->mouseDragged(x, y);
			}
			else if (event.type == SDL_QUIT)
				end();
		}

		if (isPlaying())
			SDL_Delay(50);
	}
}
