#include "PoolTable.h"
#include <string.h>

extern ImageLoader *imageLoader;

PoolTable::PoolTable(SDL_Surface *screen, PoolGame *_game)
	: Sprite(screen, IMAGE_TABLE, true), game(_game)
{
	int i;
	Ball *ball;

	for (i = 1; i < NUM_BALLS; i++)
	{
		ball = new Ball(screen, this, i);

		inPlayBalls.append(ball);
		obstacles.append(ball);
	}

	/* Setup the cueball. */
	cueBall = new Ball(screen, this, 0);
	inPlayBalls.append(cueBall);
	obstacles.append(cueBall);

	/* Load the ball shadow. */
	ballShadow = imageLoader->load(IMAGE_BALL_SHADOW, true);

	/* Setup the holes. */
	holes[0] = new Hole(getScreen(), TOP_LEFT);
	holes[1] = new Hole(getScreen(), TOP_RIGHT);
	holes[2] = new Hole(getScreen(), MIDDLE_LEFT);
	holes[3] = new Hole(getScreen(), MIDDLE_RIGHT);
	holes[4] = new Hole(getScreen(), BOTTOM_LEFT);
	holes[5] = new Hole(getScreen(), BOTTOM_RIGHT);

	for (i = 0; i < 6; i++)
		obstacles.append(holes[i]);

	/* Setup the sides. */
	sides[0] = topSide         = new Side(screen, IMAGE_TOP_SIDE);
	sides[1] = bottomSide      = new Side(screen, IMAGE_BOTTOM_SIDE);
	sides[2] = topLeftSide     = new Side(screen, IMAGE_TOPLEFT_SIDE);
	sides[3] = topRightSide    = new Side(screen, IMAGE_TOPRIGHT_SIDE);
	sides[4] = bottomLeftSide  = new Side(screen, IMAGE_BOTTOMLEFT_SIDE);
	sides[5] = bottomRightSide = new Side(screen, IMAGE_BOTTOMRIGHT_SIDE);

	for (i = 0; i < 6; i++)
		obstacles.append(sides[i]);

	moved();

	/* Set the theme. */
	setTheme(TABLE_GREEN);

	prerenderTable();
}

PoolTable::~PoolTable()
{
	inPlayBalls.clear();
	outOfPlayBalls.clear();
	obstacles.clear();
}

const char *
PoolTable::getName() const
{
	return "PoolTable";
}

void
PoolTable::setTheme(TableTheme theme)
{
	switch (theme)
	{
		case TABLE_RED:
			tableR = TABLE_RED_R;
			tableG = TABLE_RED_G;
			tableB = TABLE_RED_B;
			break;

		case TABLE_GREEN:
			tableR = TABLE_GREEN_R;
			tableG = TABLE_GREEN_G;
			tableB = TABLE_GREEN_B;
			break;

		default:
			break;
	}
}

void
PoolTable::resetCueBall()
{
	if (outOfPlayBalls.find(cueBall) != -1)
		inPlayBalls.append(outOfPlayBalls.take());

	if (obstacles.find(cueBall) == -1)
		obstacles.append(cueBall);

	/* Center the cue ball. */
	cueBall->moveTo(getX() + (getWidth() - cueBall->getWidth()) / 2 + 5,
					getY() + 235);

	renderBalls();
}

void
PoolTable::render(bool update)
{
	/* Render the table image. */
	Sprite::render(false);

	if (game->isPlaying())
	{
		Ball *ball;

		renderShadows();

		for (ball = inPlayBalls.first();
			 ball != NULL;
			 ball = inPlayBalls.next())
		{
			ball->render(false);
		}
	}

	if (update)
		SDL_UpdateRect(getScreen(), getX().intVal(), getY().intVal(),
					   getWidth().intVal(), getHeight().intVal());
}

void
PoolTable::renderShadows()
{
	Ball *ball;

	for (ball = inPlayBalls.first();
		 ball != NULL;
		 ball = inPlayBalls.next())
	{
		SDL_Rect rect;

		rect.x = ball->getX().intVal();
		rect.y = ball->getY().intVal();
		rect.w = ballShadow->w;
		rect.h = ballShadow->h;

		SDL_BlitSurface(ballShadow, NULL, getScreen(), &rect);
	}
}

void
PoolTable::renderBalls(bool update)
{
	Ball *ball;
	int i = 0;
	
	SDL_Rect *rects;

	if (update)
		rects = new SDL_Rect[inPlayBalls.count()];

	ballShadow = imageLoader->load(IMAGE_BALL_SHADOW, true);

	/* Clear the background. */
	for (ball = inPlayBalls.first(); ball != NULL; ball = inPlayBalls.next())
	{
		SDL_Rect destRect = ball->getPrevRect();

		if (update && destRect.x == 0 && destRect.y == 0)
		{
			destRect.x = ball->getX().intVal();
			destRect.y = ball->getY().intVal();
			destRect.w = ballShadow->w;
			destRect.h = ballShadow->h;
		}
		else if (destRect.x != 0 || destRect.y != 0)
		{
			destRect.w = ballShadow->w;
			destRect.h = ballShadow->h;

			render(destRect, false);
		}

		if (update)
			rects[i++] = destRect;
	}

	renderShadows();

	i = 0;

	for (ball = inPlayBalls.first(); ball != NULL; ball = inPlayBalls.next())
	{
		ball->render(false);

		if (update)
		{
			if (ball->getX() >= rects[i].x)
				rects[i].w += (ball->getX() - rects[i].x).intVal();
			else
			{
				rects[i].w += (Fixed(rects[i].x) - ball->getX()).intVal();
				rects[i].x = ball->getX().intVal();
			}
			
			if (ball->getY() >= rects[i].y)
				rects[i].h += (ball->getY() - rects[i].y).intVal();
			else
			{
				rects[i].h += (Fixed(rects[i].y) - ball->getY()).intVal();
				rects[i].y = ball->getY().intVal();
			}

			rects[i].w++;
			rects[i].h++;

			i++;
		}
	}

	if (update)
		SDL_UpdateRects(getScreen(), i, rects);

	delete [] rects;
}

const QList<Ball> &
PoolTable::getInPlayBalls() const
{
	return inPlayBalls;
}

const QList<Sprite> &
PoolTable::getObstacles() const
{
	return obstacles;
}

void
PoolTable::moved()
{
	/* Update the sides. */
	topSide->moveTo(getX()         +  30, getY() +  20);
	bottomSide->moveTo(getX()      +  30, getY() + 281);
	topLeftSide->moveTo(getX()     +  20, getY() +  30);
	topRightSide->moveTo(getX()    + 145, getY() +  30);
	bottomLeftSide->moveTo(getX()  +  20, getY() + 158);
	bottomRightSide->moveTo(getX() + 145, getY() + 158);

#ifdef NEW_SIDE_CODE
	topSide->setInfo(        0, 1, 1, 1,  0,  1);
	bottomSide->setInfo(     0, 0, 1, 0,  0, -1);
	topLeftSide->setInfo(    1, 0, 1, 1,  1,  0);
	topRightSide->setInfo(   0, 0, 0, 1, -1,  0);
	bottomLeftSide->setInfo( 1, 0, 1, 1,  1,  0);
	bottomRightSide->setInfo(0, 0, 0, 1, -1,  0);
#else
	topSide->setInfo(         0,  1, topSide->getY().neg()        - 3);
	bottomSide->setInfo(      0, -1, bottomSide->getY()           + 1);
	topLeftSide->setInfo(     1,  0, topLeftSide->getX().neg()    - 3);
	topRightSide->setInfo(   -1,  0, topRightSide->getX()         + 1);
	bottomLeftSide->setInfo(  1,  0, bottomLeftSide->getX().neg() - 3);
	bottomRightSide->setInfo(-1,  0, bottomRightSide->getX()      + 1);
#endif /* !NEW_SIDE_CODE */

	/* Update the holes. */
	holes[0]->moveTo(getX() +  13, getY() +  13);
	holes[1]->moveTo(getX() + 140, getY() +  13);
	holes[2]->moveTo(getX() +   6, getY() + 142);
	holes[3]->moveTo(getX() + 147, getY() + 142);
	holes[4]->moveTo(getX() +  13, getY() + 276);
	holes[5]->moveTo(getX() + 140, getY() + 276);
}

void
PoolTable::prerenderTable()
{
	SDL_Surface *table;
	SDL_Rect rect;
	int i;

	/* Create the table. */
	table = getImage();

	/* Fill in the background. */
	rect.x = 20;
	rect.y = 20;
	rect.w = 130;
	rect.h = 266;

	SDL_FillRect(table, &rect, SDL_MapRGB(table->format,
										  tableR, tableG, tableB));

	/* Blit the actual table on top of it. */
	SDL_BlitSurface(imageLoader->load(IMAGE_TABLE, true), NULL, table, NULL);

	/* Blit the holes and sides. */
	for (i = 0; i < 6; i++)
	{
		SDL_Rect holeRect, sideRect;

		/* Hole. */
		holeRect = holes[i]->getRect();

		SDL_BlitSurface(holes[i]->getImage(), NULL, table, &holeRect);

		/* Side. */
		sideRect = sides[i]->getRect();

		SDL_BlitSurface(sides[i]->getImage(), NULL, table, &sideRect);
	}

	/* Set the new image. */
	setImage(table);
}

bool
PoolTable::cueBallHit(CueStick *, Ball *)
{
	bool result;

	render();

	result = animLoop();

	getGame()->getRoom()->getPlayerDisplay()->render();

	return result;
}

void
PoolTable::stickMoved(CueStick *stick, Fixed oldAngle,
					  Fixed newAngle)
{
	oldAngle = newAngle = 0;

	stick->render();

//#define DRAW_GUIDELINES

#ifdef DRAW_GUIDELINES
	updateAimLines();
#endif
}

void
PoolTable::updateAimLines()
{
	static SDL_Rect rect;
	static bool firstTime = true;

	Sint16 x1, y1, x2, y2;
	Sint16 x, y, w, h;
	CueStick *stick = getGame()->getRoom()->getCueStick();
	Fixed centerX, centerY;
	Fixed distance = stick->getDistance();
	Fixed length   = 120;
	Fixed cosangle = stick->getAngleRad().cos();
	Fixed sinangle = stick->getAngleRad().sin();

	centerX = cueBall->getCenterX();
	centerY = cueBall->getCenterY();

	x1 = (centerX + sinangle * distance).intVal();
	y1 = (centerY - cosangle * distance).intVal();

	x2 = (centerX + sinangle * (distance + length).neg()).intVal();
	y2 = (centerY - cosangle * (distance + length).neg()).intVal();

	x = MIN(x1, x2);
	y = MIN(y1, y2);
	w = MAX(x1, x2) - x;
	h = MAX(y1, y2) - y;

	if (!firstTime)
		render(rect, false);

	aalineRGBA(getScreen(), x1, y1, x2, y2, 255, 255, 255, 128);

	SDL_UpdateRect(getScreen(), MIN(x, rect.x), MIN(y, rect.y),
				   MAX(w, rect.w), MAX(h, rect.h));

	/* Render the balls so it doesn't disappear on us :P */
	game->getTable()->renderBalls();

	rect.x = x;
	rect.y = y;
	rect.w = w;
	rect.h = h;

	firstTime = false;
}

bool
PoolTable::animLoop()
{
	Ball *ball, *nextBall = NULL;
	Hole *hole;
	bool animating;
	QList<Ball> inHoleBalls;
	QList<Hole> inHoles;
	bool ownBallHit = false;
	bool ballHit = false;
	bool continueTurn = false;

	do
	{
		Fixed step = 1, ballMinStep = 1, sideMinStep = 1;
		Fixed minStep = 1;
		Sprite *obstacle;

		animating = false;

		do
		{
			Ball *ballA = NULL, *ballB = NULL, *ballC = NULL, *ballD = NULL;
			Side *side = NULL;
			bool ballCollision = false;
			bool sideCollision = false;

			for (ball = inPlayBalls.first();
				 ball != NULL;
				 ball = nextBall)
			{
				if (!ball->isMoving())
				{
					nextBall = inPlayBalls.next();
					continue;
				}

				nextBall = NULL;

				for (obstacle = obstacles.first();
					 obstacle != NULL;
					 obstacle = obstacles.next())
				{
					if (ball == obstacle)
						continue;

					if (!strcmp(obstacle->getName(), "Ball"))
					{
						Ball *tempBall = (Ball *)obstacle;
						Fixed stepRet = ball->collisionPrediction(tempBall);

						if (stepRet >= 0 && stepRet < ballMinStep)
						{
							ballCollision = true;
							ballMinStep = stepRet;

							if (ballA == ball && ballB != NULL)
								ballC = tempBall;
							else
							{
								ballB = tempBall;
								ballC = NULL;
							}

							ballA = ball;
						}
					}
					else if (!strcmp(obstacle->getName(), "Side"))
					{
						Side *tempSide = (Side *)obstacle;
						Fixed stepRet =
							tempSide->collisionPrediction(ball, minStep);

						if (stepRet == 0)
							printf("Ball: %d\n", ball->getNumber());

						if (stepRet >= 0 && stepRet < sideMinStep)
						{
#if 0
							if (sideCollision)
								printf("OVERRIDING!\n");
#endif
							sideCollision = true;
							sideMinStep = stepRet;
							ballD = ball;
							side  = tempSide;
						}
					}
					else if (!strcmp(obstacle->getName(), "Hole"))
					{
						hole = (Hole *)obstacle;

						if (hole->collidesWith(ball))
						{
							SDL_Rect rect;

							inHoleBalls.append(inPlayBalls.take());

							nextBall = inPlayBalls.current();

							inHoles.append(hole);

							/* Remove from the obstacles list. */
							obstacles.remove(ball);

							/* Place the background image back on. */
							rect = ball->getRect();

							rect.w = ballShadow->w;
							rect.h = ballShadow->h;

							render(rect, true);

							break;
						}
					}
				}

				if (nextBall == NULL)
					nextBall = inPlayBalls.next();
			}

			if (sideMinStep < ballMinStep && sideMinStep < step)
				minStep = sideMinStep;
			if (ballMinStep < sideMinStep && ballMinStep < step)
				minStep = ballMinStep;

			for (ball = inPlayBalls.first();
				 ball != NULL;
				 ball = inPlayBalls.next())
			{
				if (ball->isMoving())
				{
					ball->moveBy(ball->getVectorX() * minStep,
								 ball->getVectorY() * minStep);

#if 0
					if (sideMinStep != step && minStep == sideMinStep)
					{
						renderBalls();
						game->getPhysics()->ballToSideCollision(ball, side);
						return false;
					}
#endif

					animating = true;
				}
			}

			renderBalls();

			if (ballA != NULL && ballB != NULL && ballMinStep == minStep &&
				ballCollision)
			{
				game->getPhysics()->ballToBallCollision(ballA, ballB, ballC);

				if (!ballHit && ballA == cueBall)
				{
					if (game->getRules()->ballHit(ballB) == VALID_BALL)
						ownBallHit = true;

					ballHit = true;
				}
			}

			if (ballD != NULL && side != NULL && sideMinStep == minStep &&
				sideCollision)
			{
				game->getPhysics()->ballToSideCollision(ballD, side);
			}

			for (ball = inPlayBalls.first();
				 ball != NULL;
				 ball = inPlayBalls.next())
			{
				game->getPhysics()->updateBallPos(ball, (ballD == ball));
			}

			if (!ballCollision && !sideCollision)
				break;

			step       -= minStep;
			minStep     = step;
			sideMinStep = step;
			ballMinStep = step;

		} while (step > 0 && step < 1);

		SDL_Delay(30);

		for (ball = inPlayBalls.first();
			 ball != NULL;
			 ball = inPlayBalls.next())
		{
			if (ball->isMoving())
			{
				ball->decel(0.01);
			}
		}

	} while (animating);

	Ball *firstBall;

	/* Look at the balls that were sunk in this hit. */
	for (firstBall = ball = inHoleBalls.first(), hole = inHoles.first();
		 ball != NULL && hole != NULL;
		 ball = inHoleBalls.next(), hole = inHoles.next())
	{
		BallResult result;

		inPlayBalls.remove(ball);
		outOfPlayBalls.append(ball);
#if 0
		if (inHoleBalls.find(ball) != -1)
			outOfPlayBalls.append(inHoleBalls.take());
#endif

		outOfPlayBalls.append(ball);

		result = getGame()->getRules()->ballInHole(ball, hole);

		if (result == VALID_BALL && ball == firstBall)
			continueTurn = true;

		if (result == SCRATCHED)
			resetCueBall();
		else if (result == PLAYER_LOST)
		{
			game->setWinner(game->getInactivePlayer());
			game->end();
		}
		else if (result == PLAYER_WON)
		{
			game->setWinner(game->getActivePlayer());
			game->end();
		}
	}

	return continueTurn;
}
