#include "Sprite.h"
#include <stdlib.h>

extern ImageLoader *imageLoader;

//#define BOUNDING_RECTS

Sprite::Sprite(SDL_Surface *_screen, Uint16 imgNum, bool alphaChannel)
	: x(0), y(0), animating(false), screen(_screen)
{
	image = imageLoader->load(imgNum, alphaChannel);

#ifdef BOUNDING_RECTS
	rectangleRGBA(image, 0, 0, getWidth().intVal() - 1,
				  getHeight().intVal() - 1, 0xFF, 0, 0, 0xFF);
#endif

	prevRect.x = 0;
	prevRect.y = 0;
	prevRect.w = 0;
	prevRect.h = 0;
}

Sprite::Sprite(SDL_Surface *_screen)
	: x(0), y(0), animating(false), screen(_screen)
{
	prevRect.x = 0;
	prevRect.y = 0;
	prevRect.w = 0;
	prevRect.h = 0;

	image = NULL;
}

Sprite::~Sprite()
{
}

void
Sprite::moveTo(Fixed x, Fixed y)
{
	this->x = x;
	this->y = y;

	moved();
}

void
Sprite::moveBy(Fixed x, Fixed y)
{
	moveTo(getX() + x, getY() + y);
}

void
Sprite::setX(Fixed x)
{
	moveTo(x, getY());
}

void
Sprite::setY(Fixed y)
{
	moveTo(getX(), y);
}

Fixed
Sprite::getX() const
{
	return x;
}

Fixed
Sprite::getY() const
{
	return y;
}

Fixed
Sprite::getWidth() const
{
	if (image == NULL)
		return 0;

	return Fixed(image->w);
}

Fixed
Sprite::getHeight() const
{
	if (image == NULL)
		return 0;

	return Fixed(image->h);
}

SDL_Rect
Sprite::getRect() const
{
	SDL_Rect r;

	r.x = getX().intVal();
	r.y = getY().intVal();
	r.w = getWidth().intVal();
	r.h = getHeight().intVal();

	return r;
}

bool
Sprite::containsPoint(unsigned int x, unsigned int y) const
{
	return (x >= getX().intVal() &&
			y >= getY().intVal() &&
			x < (getX() + getWidth()).intVal() &&
			y < (getY() + getHeight()).intVal());
}

void
Sprite::setAnimating(bool animating)
{
	this->animating = animating;
}

bool
Sprite::isAnimating() const
{
	return animating;
}

void
Sprite::setImage(SDL_Surface *image)
{
	if (this->image == image)
		return;

	if (this->image != NULL)
		SDL_FreeSurface(this->image);

	this->image = image;
}

SDL_Surface *
Sprite::getImage() const
{
	return image;
}

SDL_Surface *
Sprite::getScreen() const
{
	return screen;
}

bool
Sprite::collidesWith(Sprite *sprite)
{
	if (getImage() == NULL)
		return false;

	Fixed x, y, sx, sy;
	Fixed left1, left2;
	Fixed right1, right2;
	Fixed top1, top2;
	Fixed bottom1, bottom2;

	if (sprite == this)
		return false;

	x = getX();
	y = getY();
	sx = sprite->getX();
	sy = sprite->getY();

	left1 = x + getWidth()  * 0.1;
	top1  = y + getHeight() * 0.1;

	left2 = sx + sprite->getWidth()  * 0.1;
	top2  = sy + sprite->getHeight() * 0.1;

	right1  = x + getWidth()  * 0.9;
	bottom1 = y + getHeight() * 0.9;

	right2  = sx + sprite->getWidth()  * 0.9;
	bottom2 = sy + sprite->getHeight() * 0.9;

	if (bottom1 < top2) return false;
	if (top1 > bottom2) return false;
	if (right1 < left2) return false;
	if (left1 > right2) return false;

	return true;
}

void
Sprite::render(bool update)
{
	if (getImage() == NULL)
		return;

	SDL_Rect rect;

	rect.x = getX().intVal();
	rect.y = getY().intVal();
	rect.w = getWidth().intVal();
	rect.h = getHeight().intVal();

	SDL_BlitSurface(getImage(), NULL, screen, &rect);

	if (update)
		SDL_UpdateRect(getScreen(), rect.x, rect.y, rect.w, rect.h);

	prevRect = rect;
}

void
Sprite::render(SDL_Rect rect, bool update)
{
	if (getImage() == NULL)
		return;

	if (rect.x < 0)
		rect.x = 0;
	
	if (rect.y < 0)
		rect.y = 0;

	SDL_Rect destRect = rect;

	rect.x -= getX().intVal();
	rect.y -= getY().intVal();

	SDL_BlitSurface(getImage(), &rect, screen, &destRect);

	if (update)
		SDL_UpdateRect(getScreen(), destRect.x, destRect.y,
					   destRect.w, destRect.h);

	prevRect = destRect;
}

bool
Sprite::mousePressed(unsigned int, unsigned int)
{
	return false;
}

bool
Sprite::mouseReleased(unsigned int, unsigned int)
{
	return false;
}

bool
Sprite::mouseDragged(unsigned int, unsigned int)
{
	return false;
}

